"""
Collected utility functions for across Pedal.

To decide if something belongs here, ask if multiple Tools might want to use it,
possibly including Report itself. Ask if it's something you'd probably find
in a separate module, rather than being a strictly pedagogical tool. All of
these are evidence that it belongs here.

Is it something that an Instructor might want to change? Then it probably
belongs in Report or something tool specific. But you might want to abstract
away parts into here and parts into Report.
"""